<?php

namespace Drupal\tmgmt_contentapi\Services;

use Drupal\tmgmt\JobInterface;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\SourceFileApi;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\TokenApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi;
use Drupal\tmgmt_contentapi\Util\ConentApiHelper;
use Drupal\tmgmt_contentapi\Util\GeneralHelper;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateRequestFile;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\ProviderId;
use Drupal\Core\Render\Markup;
use Drupal\Component\Utility\Xss;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\Job;

/**
 * Service to create a job.
 */
class CreateConnectorJob {
  /**
   * The array for content api bundle.
   *
   * @var array
   */
  protected $contentApiBundle;

  /**
   * The contains job request object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob
   */
  protected $jobRequest;

  /**
   * The contains job api object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi
   */
  protected $jobApi;

  /**
   * The contains cp Jobs.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Model\Job
   */
  protected $createdCpJob;

  /**
   * The Source file api.
   *
   * @var useDrupal\tmgmt_contentapi\Swagger\Client\Api\SourceFileApi
   */
  protected $fileApi;

  /**
   * The request api.
   *
   * @var useDrupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi
   */
  protected $transRequestApi;

  /**
   * The transfer file service.
   *
   * @var \Drupal\tmgmt_contentapi\Services\ExportJobFiles
   */
  protected $transferFiles;

  /**
   * Token to submit the job.
   *
   * @var array
   */
  protected $token;

  /**
   * Token API object.
   *
   * @var array
   */
  protected $capi;

  /**
   * The stream wrapper service.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManager
   */
  protected $streamWrapperManager;

  /**
   * The filesystem service.
   *
   * @var \Drupal\Core\Datetime\DrupalDateTime
   */
  protected $drupalDateTime;

  /**
   * Message to pass when job created.
   *
   * @var string
   */
  protected $messageToPass;

  /**
   * Consructor.
   *
   * @param ExportJobFiles $transferFiles
   *   Object for custom service export job files.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $streamWrapperManager
   *   Object for stream wrapper manager service.
   */
  public function __construct(
    ExportJobFiles $transferFiles,
    StreamWrapperManagerInterface $streamWrapperManager
    ) {
    $this->transferFiles = $transferFiles;
    $this->streamWrapperManager = $streamWrapperManager;
    // Set default values for variables.
    $this->contentApiBundle = [];
    $this->jobRequest = NULL;
    $this->jobApi = new JobApi();
    $this->fileApi = '';
    $this->transRequestApi = '';
    $this->capi = '';
    $this->token = '';
    $this->messageToPass = 'Exported files can be downloaded here: <br/>';

  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('tmgmt_contentapi.export_job'),
      $container->get('stream_wrapper_manager'),

    );
  }

  /**
   * Function to set contentApiBundle.
   *
   * @return array
   *   content api bundle
   */
  public function setContentApiBundle(array $value) {
    return $this->contentApiBundle = $value;

  }

  /**
   * Function to set jobRequest.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateJob $value
   *   Set jobRequest object.
   */
  public function setJobRequest(CreateJob $value) {
    return $this->jobRequest = $value;

  }

  /**
   * Function to set jobApi.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi $value
   *   Set job api.
   */
  public function setJobApi(JobApi $value) {
    return $this->jobApi = $value;
  }

  /**
   * Function to set createdCpJob.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Model\Job $value
   *   Set job api.
   */
  public function setCreatedCpJob(Job $value) {
    return $this->createdCpJob = $value;

  }

  /**
   * Function to set fileApi.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Api\SourceFileApi $value
   *   Set file api variable.
   */
  public function setFileApi(SourceFileApi $value) {
    return $this->fileApi = $value;
  }

  /**
   * Function to set transRequestApi.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi $value
   *   Set tansaction request api.
   */
  public function setTransRequestApi(RequestApi $value) {
    return $this->transRequestApi = $value;
  }

  /**
   * Function to set capi.
   *
   * @param \Drupal\tmgmt_contentapi\Swagger\Client\Api\TokenApi $value
   *   Set capi token.
   */
  public function setCapi(TokenApi $value) {
    return $this->capi = $value;
  }

  /**
   * Function to set token.
   *
   * @param string $value
   *   Token for processing jobs.
   */
  public function setToken(string $value) {
    return $this->token = $value;
  }

  /**
   * Function to set drupal date time.
   *
   * @param \Drupal\Core\Datetime\DrupalDateTime $value
   *   Set drupal date and time object.
   */
  public function setDrupalDateTime(DrupalDateTime $value) {
    return $this->drupalDateTime = $value;
  }

  /**
   * Function to get job files.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   *
   * @return array
   *   job file to transfer
   */
  public function getJobFilesToTransfer(JobInterface $job) {
    return $this->transferFiles->jobFilesToTransfer($job);
  }

  /**
   * Function to get reference files.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   *
   * @return array
   *   reference files to transfer
   */
  public function getRefFilesToTransfer(JobInterface $job) {
    return $this->transferFiles->refFilesToTransfer($job);
  }

  /**
   * Function to create cp jobs.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function createCpJobs(JobInterface $job) {
    // Create cp job.
    $this->setJobRequest(ConentApiHelper::genrateJobRequst($job));
    $translator = $job->getTranslator();
    $capisettings = $translator->getSetting('capi-settings');
    $this->setCapi(new TokenApi());

    $this->setToken($this->capi->getToken($capisettings['capi_username_ctt'], $capisettings['capi_password_ctt']));

    try {
      $this->createdCpJob = $this->jobApi->jobsPost($this->token, $this->jobRequest);
    }
    catch (\Exception $e) {
      $msg = $e->getMessage();
      if (strlen($msg) > 200) {
        $msg = substr($msg, 0, 200);
      }
      $job->addMessage($msg, [], 'error');
      return;
    }
  }

  /**
   * Function to transfer files to cpa.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   */
  public function uploadTransferFilesToCpa(JobInterface $job) {
    $transfer_files = $this->getJobFilesToTransfer($job);
    $this->createCpJobs($job);
    // Upload files from transferfiles array to CPA.
    $this->setFileApi(new SourceFileApi());
    $this->setTransRequestApi(new RequestApi());
    // Array to store temporally translation requst and associated files.
    $contentapitrrequstfiles = [];
    foreach ($transfer_files as $id => $tmpfile) {
      $data = [];
      $data["job_id"] = $this->createdCpJob->getJobId();
      $data["filename"] = $tmpfile->getFilename();
      $stmrg = $this->streamWrapperManager->getViaUri($tmpfile->getFileUri());
      $extpath = $stmrg->realpath();
      $zip = new \ZipArchive();
      if ($zip->open($extpath) === TRUE) {
        $zip->extractTo($this->transferFiles->allFilesPath);
        $zip->close();
      }
      $data["filetype"] = $tmpfile->getMimeType();
      $stmrg = $this->streamWrapperManager->getViaUri($tmpfile->getFileUri());
      $extpath = $stmrg->realpath();
      $filrequst = new \SplFileObject($extpath);
      // File upload to TUS CLINET.
      $name = $data["filename"];
      $jobpath = $job->getTranslator()->getSetting('transfer-settings') ? $extpath : $this->transferFiles->allFilesPath . "/" . $name;
      $contentapitmpfile = $this->fileApi->jobsJobIdUploadPost($this->token, $data["job_id"], $data["filename"], $data["filetype"], $filrequst);

      // $tus_client = $fileapi->tusClient($contentapitmpfile, $jobpath);
      $uploadedFile = $this->fileApi->uploadFile($contentapitmpfile, $jobpath);

      $request = new CreateRequestFile([
        "request_name" => $job->id() . "_" . $id . "_" . $job->getRemoteSourceLanguage() . "-" . $job->getRemoteTargetLanguage(),
        "source_native_id" => $job->id() . "_" . $id,
        "source_native_language_code" => $job->getRemoteSourceLanguage(),
        "target_native_language_codes" => [$job->getRemoteTargetLanguage()],
        "file_id" => $contentapitmpfile['fms_file_id'],
      ]);
      $contentapitmprequst = $this->transRequestApi->jobsJobIdRequestsAddfilePost($this->token, $data["job_id"], $request);
      $this->contentApiBundle[] = $contentapitmprequst;
    }
    // Send reference files with the job.
    $this->getRefFilesToTransfer($job);
  }

  /**
   * Function to submit job to content API.
   *
   * @param \Drupal\tmgmt\JobInterface $job
   *   Job under process.
   *
   * @return string
   *   confirmation messsage with job details
   */
  public function submitJobToContentApi(JobInterface $job) {
    $this->setDrupalDateTime(new DrupalDateTime('now', 'UTC'));
    $this->uploadTransferFilesToCpa($job);
    // Submit content api job.
    $pr = $job->getSetting("capi-settings")["provider"];
    $jid = $this->createdCpJob->getJobId();
    $prmodel = new ProviderId(['provider_id' => $pr]);
    $jobsubmitresult = $this->jobApi->jobsJobIdSubmitPut($this->token, $jid, $prmodel);
    if ($jobsubmitresult) {
      // TODO
      // Add content api bundle to job's settings.
      GeneralHelper::addCpaSettingsToJob($job, serialize($this->contentApiBundle));
      // Generate message which user can see once job created.
      $this->messageToPass .= '<a href="' . \Drupal::service('file_url_generator')->generateAbsoluteString($this->transferFiles->zipPath) . '">' . Xss::filter(GeneralHelper::getJobLabelNoSpeChars($job)) . '</a>';
      \Drupal::messenger()->addMessage(Markup::create($this->messageToPass));
      // Set message if quote is set else success.
      if ($job->getSetting('capi-settings')['quote']['is_quote']) {
        $job->submitted("This job was submitted for a quote. To submit your job for processing, you must log into your translation provider's system to approve this quote.");
      }
      else {
        $job->submitted("Job sent to provider!");
      }

      // Delete files after sending to Provider.
      $this->transferFiles->deleteProcessedFiles();

      // Add job details to drupal db for tracking.
      ConentApiHelper::capiJobSetStoredData(
        $job,
        [
          'jobId' => $jobsubmitresult->getJobId(),
          'jobStatus' => $jobsubmitresult->getStatusCode(),
          'providerId' => $jobsubmitresult->getProviderId(),
          'jobType' => 2,
          'updateTime' => $this->drupalDateTime->format("Y-m-d H:i:s:u"),
        ]
      );
    }
    return $this->messageToPass;
  }

}
