<?php

namespace Drupal\tmgmt_contentapi\Services;

use Drupal\tmgmt_contentapi\Swagger\Client\Api\StatusUpdateApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\TokenApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\CreateToken;
use Drupal\tmgmt_contentapi\Util\ConentApiHelper;
use Drupal\tmgmt\Entity\Job;
use Drupal\tmgmt\Entity\JobItem;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\FileApi;
use Drupal\tmgmt_contentapi\Swagger\Client\Model\ArrayOfRequestIds;
use Drupal\tmgmt_contentapi\Swagger\Client\Configuration;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\File\FileRepositoryInterface;
use Symfony\Component\Mime\MimeTypeGuesserInterface;
use Drupal\file\Entity\File;
use Drupal\tmgmt_contentapi\Plugin\tmgmt_contentapi\Format\Xliff;
use Drupal\tmgmt_contentapi\Util\GeneralHelper;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\tmgmt\TranslatorInterface;
use Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\SuspendQueueException;
use \Drupal\Core\Utility\Error;



/**
 * Service to process job.
 *
 * Acknowledge, update, delete jobs.
 *
 * @category Class
 * @package Drupal\tmgmt_contentapi\Services
 * @author Lionbridge team
 */
class ProcessJob {
  /**
   * The job id which nees to processed.
   *
   * @var string
   */
  protected $filterJobId;

  /**
   * The capi jobs array.
   *
   * @var array
   */
  protected $capiJobs;

  /**
   * The ctt jobs array.
   *
   * @var array
   */
  protected $jobsCtt;

  /**
   * Status api object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\StatusUpdateApi
   */
  protected $statusApi;

  /**
   * The capi v1 jobs array.
   *
   * @var array
   */
  protected $capiV1Jobs;

  /**
   * The capi v2 jobs array.
   *
   * @var array
   */
  protected $capiV2Jobs;

  /**
   * The capi token.
   *
   * @var string
   */
  protected $capiToken;

  /**
   * The ctt token.
   *
   * @var string
   */
  protected $cttToken;

  /**
   * The translator object.
   *
   * @var \Drupal\tmgmt\TranslatorInterface
   */
  protected $translator;

  /**
   * Token API object.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\TokenApi
   */
  protected $capi;

  /**
   * The capi job which are ready for deliver.
   *
   * @var array
   */
  protected $capiToDeliver;

  /**
   * The capi job with status cancelled.
   *
   * @var array
   */
  protected $capiCancelled;

  /**
   * The capi job which are not yet ready for deliver.
   *
   * @var array
   */
  protected $capiToIgnore;

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\RequestApi
   */
  protected $requestApi;

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\FileApi
   */
  protected $fileApi;

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Model\ArrayOfRequestIds
   */
  protected $arrayOfRequestId;

  /**
   * Public file path for received files.
   */
  const PUBLIC_RECEIVED_FILE_PATH = '://tmgmt_contentapi/LioxReceivedFiles/';

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Configuration
   */
  protected $config;

  /**
   * The filesystem service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The file repository service.
   *
   * @var \Drupal\File\FileRepositoryInterface
   */
  protected $fileRepository;

  /**
   * The stream wrapper service.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * If provider allows to auto accept jobs.
   *
   * @var bool
   */
  protected $isAutoAccept;

  /**
   * Check if auto import is on.
   *
   * @var bool
   */
  protected $isAutoImport;

  /**
   * The stream wrapper service.
   *
   * @var \Symfony\Component\Mime\MimeTypeGuesserInterface
   */
  protected $guesser;

  /**
   * The request api.
   *
   * @var \Drupal\tmgmt_contentapi\Swagger\Client\Api\JobApi
   */
  protected $jobApi;

  /**
   * The filesystem service.
   *
   * @var \Drupal\Core\Datetime\DrupalDateTime
   */
  protected $drupalDateTime;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * The queue object used for processing jobs.
   *
   * @var object
   */
  protected $queue;

  /**
   * The queue object used for processing jobs.
   *
   * @var int
   */
  protected $totalDeliverables;

  /**
   * Consructor.
   *
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   Object for file system service.
   * @param \Drupal\File\FileRepositoryInterface $fileRepository
   *   Object for file repository service.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $streamWrapperManager
   *   Object for stream wrapper manager service.
   * @param \Symfony\Component\Mime\MimeTypeGuesserInterface $guesser
   *   Object for mime type guesser.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   Service for queue factory.
   */
  public function __construct(
    FileSystemInterface $fileSystem,
    FileRepositoryInterface $fileRepository,
    StreamWrapperManagerInterface $streamWrapperManager,
    MimeTypeGuesserInterface $guesser,
    QueueFactory $queue_factory

    ) {
      $this->fileSystem = $fileSystem;
      $this->fileRepository = $fileRepository;
      $this->streamWrapperManager = $streamWrapperManager;
      $this->guesser = $guesser;
      $this->queueFactory = $queue_factory;
      // Set default values for variables.
      $this->queue = $this->generateQueueObject('import_translated_jobs_from_queue');
      $this->filterJobId = '';
      $this->capiJobs = [];
      $this->jobsCtt = [];
      $this->capi = new TokenApi();
      $this->statusApi = new StatusUpdateApi();
      $this->requestApi = new RequestApi();
      $this->fileApi = new fileApi();
      $this->jobApi = new JobApi();
      $this->config = new Configuration();
      $this->drupalDateTime = new DrupalDateTime('now', 'UTC');
      $this->translator = $this->setTranslator();
      $this->capiToken = $this->getCapiToken();
      $this->cttToken = "";//$this->getCttToken();
      $this->isAutoAccept = $this->translator->isAutoAccept();
      $this->isAutoImport = $this->isAutoImportSet();
      $this->capiV1Jobs = [];
      $this->capiToDeliver = [];
      $this->capiCancelled = [];
      $this->capiToIgnore = [];
      $this->totalDeliverables = 0;
      //if (!empty($this->cttToken))
      //{
      //  $this->capiV1Jobs = $this->statusApi->jobsJobIdStatusupdatesGet($this->cttToken);
      //}
      $this->capiV2Jobs = $this->statusApi->statusupdatesGet($this->capiToken);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('file.system'),
      $container->get('file.repository'),
      $container->get('stream_wrapper_manager'),
      $container->get('file.mime_type.guesser'),
      $container->get('queue'),
    );
  }

  function generateQueueObject($queue_name) {
    $queue = \Drupal::queue($queue_name);
    return $queue;
  }

  /**
   * Function to set capi V2 Jobs.
   *
   * @param string $value
   *   Specific job id for porcessing.
   */
  protected function setFilterJobId(string $value) {
    return $this->filterJobId = $value;
  }

  /**
   * Function to set set translator.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   return translator object
   */
  protected function setTranslator() {
    // Get all translators.
    $translators = \Drupal::entityTypeManager()->getStorage('tmgmt_translator')->loadByProperties(['plugin' => 'contentapi']);
    if (!$translators) {
      return [];
    }
    foreach ($translators as $translator) {
      return $translator;
    }

  }

  /**
   * Function to check if auto import set.
   *
   * @return string
   *   capi token
   */
  protected function isAutoImportSet() {
    $config = $this->translator->getSetting('cron-settings');
    return $this->isAutoImport = $config['status'];
  }

  /**
   * Function to get capi token.
   *
   * @return string
   *   capi token
   */
  protected function getCapiToken() {
    $capisettings = $this->translator->getSetting('capi-settings');
    return $this->capi->getToken($capisettings['capi_username_ctt'], $capisettings['capi_password_ctt']);
  }

  /**
   * Function to get ctt token.
   *
   * @return string
   *   ctt token
   */
  protected function getCttToken() {
    $capisettings = $this->translator->getSetting('capi-settings');
    if ($capisettings['capi_username_ctt1'] == '') {
      return '';
    }
    $tokenrequest = new CreateToken(
      [
        'username' => $capisettings['capi_username_ctt1'],
        'password' => $capisettings['capi_password_ctt1']
      ]
    );
    $response = $this->capi->oauth2TokenPost($tokenrequest);
    if (isset($response) != TRUE) {
      \Drupal::logger('TMGMT_CONTENTAPI')->warning("Invalid Content API V1 credentials found while running CRON");
      return '';
    }
    return $response['access_token'];
  }

  /**
   * Function to set scan the updated jobs.
   *
   * @return array
   *   capi jobs of version 1
   */
  protected function scanV1Job() {
    $jobsCtt = [];
    foreach ($this->capiV1Jobs as $value) {
      // Extract each item and put it inside $capiJobs.
      if ($this->filterJobId != '' && $this->filterJobId != $value['job_id']) {
        continue;
      }
      // Create job info array.
      $jobInfo = [];
      $jobInfo['type'] = 1;
      $jobInfo['jobId'] = $value['job_id'];
      $jobInfo['updateId'] = $value['update_id'];
      $jobInfo['requestIds'] = $value['request_ids'];
      $jobInfo['updateTime'] = $value['update_time'];
      $jobInfo['status'] = $value->getStatusCode()->getStatusCode();
      if (!array_key_exists($value['job_id'], $jobsCtt)) {
        $jobsCtt[$value['job_id']] = [];
        $jobsCtt[$value['job_id']][] = $jobInfo;
        continue;
      }
      if ($jobsCtt[$value['job_id']][0]['updateTime'] < $jobInfo['updateTime']) {
        array_unshift($jobsCtt[$value['job_id']], $jobInfo);
      }
      else {
        $jobsCtt[$value['job_id']][] = $jobInfo;
      }
    }
    foreach ($jobsCtt as $value) {
      for ($idx = 0; $idx < count($value); $idx++) {
        if ($idx == 0) {
          $this->capiJobs[$value[$idx]['jobId']] = $value[$idx];
          continue;
        }
        $this->statusApi->statusupdatesUpdateIdAcknowledgePutCtt($this->cttToken, $value[$idx]['updateId']);
      }
    }
    return $this->capiJobs;
  }

  /**
   * Function to set scan the updated jobs.
   *
   * @return array
   *   capi jobs of version 2
   */
  protected function scanV2Job() {
    $jobs = [];
    foreach ($this->capiV2Jobs as $value) {
      if ($this->filterJobId != '' && $this->filterJobId != $value['job_id']) {
        continue;
      }
      $jobInfo = [];
      $jobInfo['type'] = 2;
      $jobInfo['jobId'] = $value['job_id'];
      $jobInfo['updateId'] = $value['update_id'];
      $jobInfo['requestIds'] = $value['request_ids'];
      $jobInfo['updateTime'] = $value['update_time'];
      $jobInfo['status'] = $value['status_code'];
      $jobInfo['prevStatus'] = '';
      // Add a job into the array only if it is not already present.
      if (!array_key_exists($value['job_id'], $jobs)) {
        $jobs[$value['job_id']] = [];
        $jobs[$value['job_id']][] = $jobInfo;
        continue;
      }
      // If status update time is greater than the previous status update time then add it to the beginning of the array.
      if ($jobs[$value['job_id']][0]['updateTime'] < $jobInfo['updateTime']) {
        array_unshift($jobs[$value['job_id']], $jobInfo);
      }
      else {
        $jobs[$value['job_id']][] = $jobInfo;
      }
    }
    // Put only latest status update in the capiJobs array. Acknowledge all other status updates.
    foreach ($jobs as $value) {
      for ($idx = 0; $idx < count($value); $idx++) {
        if ($idx == 0) {
          $this->capiJobs[$value[$idx]['jobId']] = $value[$idx];
          continue;
        }
        $this->statusApi->statusupdatesUpdateIdAcknowledgePut($this->capiToken, $value[$idx]['updateId']);
      }
    }
    return $this->capiJobs;
  }

  /**
   * Load capi jobs and update jobs current status and state on drupal side.
   *
   * @return array
   *   capi jobs with the updated status
   */
  protected function updateJobStatusToLatest() {
    $processedResults = ConentApiHelper::fetchCurrentStoredJobInfo($this->capiJobs);
    foreach ($processedResults as $value) {
      $loadedData = json_decode($value['data'], TRUE);
      if (array_key_exists($loadedData['capiJob']['jobId'], $this->capiJobs) == FALSE) {
        continue;
      }
      $tmpItem = $this->capiJobs[$loadedData['capiJob']['jobId']];
      $tmpItem['providerId'] = $loadedData['capiJob']['providerId'];
      $tmpItem['prevStatus'] = $loadedData['capiJob']['status'];
      // JIRA 1194821: Job status column ambiguity issue.
      $updateTime = $tmpItem['updateTime']->format('Y-m-d H:i:s:u');
      $updateTimeInDB = $loadedData['capiJob']['updateTime'];

      $tmpItem['drupalJob'] = Job::load($value['tjid']);

      if ($tmpItem['prevStatus'] != $tmpItem['status']) {
        if ($tmpItem['prevStatus'] == 'TRANSLATION_APPROVED' && $tmpItem['status'] == 'REVIEW_TRANSLATION') {
          $tmpItem["drupalJob"]->setState(1);
        }
        // Once job is already in REVIEW_TRANSLATION and got some old status skip DB update for the same.
        if ($updateTime > $updateTimeInDB) {
          ConentApiHelper::capiJobSetStoredData(
            $tmpItem['drupalJob'],
            [
              'jobId' => $tmpItem['jobId'],
              'jobStatus' => $tmpItem['status'],
              'providerId' => $tmpItem['providerId'],
              'jobType' => $tmpItem['type'],
              'updateTime' => $updateTime
            ]
          );
        }
      }
      $this->capiJobs[$loadedData['capiJob']['jobId']] = $tmpItem;
    }
    for ($idx = count($this->capiJobs); $idx > 0; $idx--) {
      if (array_key_exists('drupalJob', $this->capiJobs[array_keys($this->capiJobs)[$idx - 1]]) == FALSE) {
        array_splice($this->capiJobs, $idx - 1, 1);
      }
    }
    return $this->capiJobs;
  }

  /**
   * Function to set scan the updated jobs.
   *
   * @param string $jobId
   *   Specific job id which needs to process else it will process all updated jobs.
   */
  protected function scanJob(string $jobId) {
    $this->setFilterJobId($jobId);
    $scan_v1_jobs = [];
    $scan_v2_jobs = [];
    if (!empty($this->capiV1Jobs)) {
      $scan_v1_jobs = $this->scanV1Job();
    }
    if (!empty($this->capiV2Jobs)) {
      $scan_v2_jobs = $this->scanV2Job();
    }
    if (empty($this->capiJobs)) {
      return;
    }
    $this->updateJobStatusToLatest();
  }

  /**
   * Function to filter and aknowledge job status.
   *
   * @param string $jobId
   *   Specific job id which needs to process else it will process all updated jobs.
   *
   * @return array
   *   capi jobs for status updation
   */
  public function filterAknowledgeDownloadJobs(string $initJobId) {
    $uniqueJobs = [];
    //$toProcessRequests = [];
    if(isset($initJobId) && $initJobId != ''){
      $uniqueJobs[] = $initJobId;
    }else{
      foreach ($this->capiV2Jobs as $value) {
        $jobId = $value["job_id"];
        if(!in_array($jobId,$uniqueJobs,false)){
          $uniqueJobs[] = $jobId;
        }
      }
    }
    if(count($uniqueJobs) == 0)
      return;
    $existingJobs = ConentApiHelper::fetchCurrentStoredJobInfoNew($uniqueJobs);
    if(!isset($initJobId) || $initJobId == ''){
      foreach ($this->capiV2Jobs as $value) {
        if(array_key_exists($value["job_id"],$existingJobs)){
          $this->statusApi->statusupdatesUpdateIdAcknowledgePut($this->capiToken, $value["update_id"]);
        }
      }
    }

    foreach($existingJobs as $value){
      $drupalJob = Job::load($value['tjid']);
      $jobInfo = json_decode($value["data"]);
      $requests = $this->requestApi->jobsJobIdRequestsGet($this->getCapiToken(), $jobInfo->capiJob->jobId);
      $updateTimeD =  new DrupalDateTime('now','UTC');
      $updateTime = $updateTimeD->format("Y-m-d H:i:s:u");
      $currStatus = $jobInfo->capiJob->status;
      foreach($requests as $req)
      {
        if($req->getStatusCode() == "REVIEW_TRANSLATION" || ($initJobId != '' && ($req->getStatusCode() == "TRANSLATION_APPROVED" || $req->getStatusCode() == "COMPLETED")))
        {
          if($drupalJob->getState() == Job::STATE_FINISHED)
          {
            $drupalJob->setState(Job::STATE_ACTIVE);
          }
          $data = [
              'job_id' =>  $jobInfo->capiJob->jobId,
              'request_id' => $req->getRequestId(),
              'tjid' => $value['tjid'],
              'context' => &$context,
          ];
          $this->queue->createItem($data);
        }
        if($req->getStatusCode() == "CANCELLED")
        {
          if($drupalJob->getState() != Job::STATE_ABORTED)
          {
            $drupalJob->setState(Job::STATE_ABORTED);
            $items = $drupalJob->getItems();
            foreach($items as $item)
            {
             $item->setState(JobItem::STATE_ABORTED);
            }
          }
        }
        $updateTime =  $req->getModifiedDate()->format("Y-m-d H:i:s:u");
        $currStatus = $req->getStatusCode();
      }
      $element_array = [
        'jobId' => $jobInfo->capiJob->jobId ,
        'jobStatus' => $currStatus,
        'providerId' => $drupalJob->getSetting("capi-settings")["provider"],
        'jobType' => 2,
        'updateTime' => $updateTime
      ];
      ConentApiHelper::capiJobSetStoredData($drupalJob, $element_array);
    }
    if($this->queue->numberOfItems() > 0){
      $batch = [
        'title' => t('Started download and import for ready jobs'),
        'operations' => [],
        'finished' => [ProcessJob::class, 'batchSubmitFinished'],
      ];

      $batch['operations'][] = [
        [ProcessJob::class, 'batchSubmit'],
        [$this->queue->numberOfItems()],
      ];
      // Initiate batch process.
      batch_set($batch);
    }
    //HERE WE ARE GOING TO GET EVERY REQUEST FROM EVERY JOB AND ENQUEUE THOSE READY
  }
  public function filterAknowledgeDownloadJobsOld(string $jobId) {
    // This functon will filter the job on the basis of current status.
    // Divide jobs into three different array.
    // Deliver: Jobs which are in state of review_translation/trasnlation_approved.
    // Cancelled: Jobs which are in cancelled state and no further processing required.
    // Ignore: Job which are in other than review_translation/translation_approved/cancelled state.
    // Ignore jobs not required to process, acknowledge the job and update job status in DB.

    // Step 1: Scan job for this provider.
    $this->scanJob($jobId);
    // If no job to aknowledge or process then return.
    if (empty($this->capiJobs)) {
      return [];
    }

    // Step 2: Divide job into deliver, cancelled or ignore status.
    $job_to_process = 0;
    foreach ($this->capiJobs as $value) {
      // If the job is not active then ignore it for processing.
      if ($value["drupalJob"]->getState() != 1) {
        $this->capiToIgnore[$value['jobId']] = $value;
        continue;
      }
      switch ($value['status']) {
        case "REVIEW_TRANSLATION":
        case "TRANSLATION_APPROVED":
          $this->capiToDeliver[$value['jobId']] = $value;
          break;

        case "CANCELLED":
          $this->capiCancelled[$value['jobId']] = $value;
          break;

        default:
          $this->capiToIgnore[$value['jobId']] = $value;
      }
    }

    foreach ($this->capiToDeliver as $value) {
      $this->totalDeliverables = $this->totalDeliverables + count($value['requestIds']);
    }

    // Step 3: Aknowledge jobs for all statuses.
    // 1: Delivered job .ie download and import the job.
    // 2. Cancelled job.
    // 0. Ignored job .ie see the latest status and aknowledge it.
    // Aknowledge Jobs which are not ready to deliver.
    $this->aknowledgeJobs($this->capiToIgnore, 0);
    // Aknowledge Jobs which are ready to deliver.
    $this->aknowledgeJobs($this->capiToDeliver, 1);
    // Aknowledge Jobs which have cancelled status.
    $this->aknowledgeJobs($this->capiCancelled, 2);

  }

  /**
   * Function to aknowledge jobs.
   *
   * @param array $capi_jobs_to_aknowledge
   *   Jobs which nees to aknowledge.
   * @param int $job_to_process
   *   Job bifurcation id.
   */
  protected function aknowledgeJobs(array $capi_jobs_to_aknowledge, int $job_to_process) {
    if (empty($capi_jobs_to_aknowledge)) {
      return;
    }
    foreach ($capi_jobs_to_aknowledge as $value) {
      // If update ID is not set then continue with next values.
      if ($value['updateId'] == '') {
        continue;
      }
      // In case of CANCELLED Job.
      if ($job_to_process == 2) {
        // THIS IS TMGMT JOB CANCEL there is no cancel in tmgmt.
        $value['drupalJob']->setState(Job::STATE_ABORTED);
        $this->sendAcknowlegement($value['type'], $value['updateId']);
      }
      // In case job needs only akowledge.
      if ($job_to_process == 0) {
        $this->sendAcknowlegement($value['type'], $value['updateId']);
      }
      // In case of job ready for delivery.
      if ($job_to_process == 1) {
        // Step 4: Job which are ready for delivery needs to donwload.
        $this->downloadJob($value);
        // Aknowledge update when all items are processed for this perticular updateid.
        // Same has been implemented in queue processing.
      }
    }
  }

  /**
   * Function to acknowledge the job.
   *
   * @param int $type
   *   Is capi version 1 or 2.
   *
   * @param string $updateId
   *   Update id which needs to acknowledge.
   */
  protected function sendAcknowlegement($type, $updateId) {
    // Aknowledge Job.
    try {
      if ($type == 1) {
        $this->statusApi->statusupdatesUpdateIdAcknowledgePutCtt($this->cttToken, $updateId);
      }
      else {
        $this->statusApi->statusupdatesUpdateIdAcknowledgePut($this->capiToken, $updateId);
      }
    }
    catch (\Exception $exception) {
      \Drupal::logger('Job acknowledge')->notice('Job already aknowledged %provider: %message', [
        '%provider' => $this->translator->id(),
        '%message' => $exception->getMessage(),
      ]);
    }
  }

  /**
   * Function to download Job, auto accept and approve.
   *
   * @param array $job_info
   *   Jobs which nees to download, autoaccept and approve.
   */
  protected function downloadJob(array $job_info) {
    try {
      $job = $job_info['drupalJob'];
      $request_ids = $job_info['requestIds'];
      $dir = $job->getSetting('scheme') . self::PUBLIC_RECEIVED_FILE_PATH;
      // Check if you able to create directory for above mentioned path.
      if (!$this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY)) {
        \Drupal::messenger()->addMessage(t('Directory could not be created: ') . $dir);
        return;
      }

      $stmrg = $this->streamWrapperManager->getViaUri($dir);
      $tempfolder = $stmrg->realpath();
      $this->config->setTempFolderPath($tempfolder);
      Configuration::setDefaultConfiguration($this->config);

      foreach ($request_ids as $request_id) {
        $data = [
          'idarray' => $request_id,
          'job_info' => $job_info,
          'context' => &$context,
        ];
        $this->queue->createItem($data);
      }

      $batch = [
        'title' => t('Started download and import for ready jobs'),
        'operations' => [],
        'finished' => [ProcessJob::class, 'batchSubmitFinished'],
      ];

      $batch['operations'][] = [
        [ProcessJob::class, 'batchSubmit'],
        [$this->queue->numberOfItems()],
      ];
      // Initiate batch process.
      batch_set($batch);

    }
    catch (\Exception $exception) {
      $respbody = $exception->getMessage();
      if ($exception instanceof ApiException) {
        $respbody = $exception->getResponseBody();
      }
      $respbody = 'An error occured while fetching and importing files: ' . $respbody;
      if (strlen($respbody) > 200) {
        $respbody = substr($respbody, 0, 200);
      }
      \Drupal::messenger()->addMessage($respbody, 'error');
    }
  }

  /**
   * Batch dispatch callback for submitting a job.
   *
   * @param int $total
   *   The job items which are processed.
   * @param array $context
   *   Job detail whose items ha been procesed.
   */
  public static function batchSubmit($total, &$context) {
    $queue_name = 'import_translated_jobs_from_queue';
     /** @var \Drupal\Core\Queue\QueueWorkerInterface $queue_worker */
    $queue_worker = \Drupal::service('plugin.manager.queue_worker')->createInstance($queue_name);
    $queue = \Drupal::queue($queue_name);

    if (!isset($context['sandbox']['progress'])) {
      $context['sandbox']['progress'] = 0;
      $context['sandbox']['max'] = $total;
    }

    while ($context['sandbox']['progress'] < $context['sandbox']['max']) {
      if ($item = $queue->claimItem()) {
        try {
          $item->data["context"] = $context;
          $queue_worker->processItem($item->data);
          $queue->deleteItem($item);
          \Drupal::logger('Item process via batch')->notice('Item Batch processing completed successfully.');
        }
        catch (SuspendQueueException $e) {
          $queue->releaseItem($item);
          break;
        }
        catch (\Exception $e) {
          $logger = \Drupal::logger('Queue');
          Error::logException($logger, $e, "Something went wrong while processing the queue item.");
          $queue->releaseItem($item);
        }
      }

      $context['sandbox']['progress']++;
      if ($context['sandbox']['progress'] >= $context['sandbox']['max']) {
        $context['finished'] = 1;
      }
      else {
        $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
      }
    }
  }
/**
   * Function to process request in a batch.
   *
   * @param string $request_id
   *   Request id to process.
   *
   * @param array $job_info
   *   Job related information.
   *
   * @param mixed string $context
   *   Context.
   */
  public function processRequestIdNew($jobId, $requestId, $drupalJob, &$context) {
    // Generate Job and item details.

    // Perform operation for request id, i.e download and import.
    $spfilearray = $this->fileApi->jobsJobIdRequestsRequestIdRetrievefileGetWithHttpInfo($this->capiToken, $jobId, $requestId);
    $spfile = $spfilearray[0];
    $fcontent = $spfile->fread($spfile->getSize());
    $fpath = $spfile->getRealPath();
    $spfilearray[0] = NULL;
    $spfile = NULL;
    $this->fileSystem->delete($fpath);
    $filename = str_replace('"', '', explode("=", explode(";", $spfilearray[2]['Content-Disposition'][0])[1])[1]);
    // Step 5: Import the job for review and accept.
    $this->importJob($drupalJob, $fcontent, $filename);

    $arrayrequestid = new ArrayOfRequestIds();
    $arrayrequestid->setRequestIds([$requestId]);
    $this->requestApi->jobsJobIdRequestsApprovePut($this->capiToken, $jobId, $arrayrequestid);

    if(!array_key_exists($jobId, $context['results'])){
      $context['results'][$jobId] = $drupalJob->id();
    }
  }
  /**
   * Function to process request in a batch.
   *
   * @param string $request_id
   *   Request id to process.
   *
   * @param array $job_info
   *   Job related information.
   *
   * @param mixed string $context
   *   Context.
   */
  public function processRequestId($request_id, $job_info, &$context) {
    // Generate Job and item details.
    $job = $job_info['drupalJob'];
    $prId = $job_info['jobId'];
    $job_from = $job_info['type'];
    $providerId = $job_info['providerId'];
    $updateId = $job_info['updateId'];

    // Perform operation for request id, i.e download and import.
    if ($job_from == 1) {
      $spfilearray = $this->fileApi->jobsJobIdRequestsRequestIdRetrievefileGetWithHttpInfoCtt($this->cttToken, $prId, $request_id);
    }
    else {
      $spfilearray = $this->fileApi->jobsJobIdRequestsRequestIdRetrievefileGetWithHttpInfo($this->capiToken, $prId, $request_id);
    }
    $spfile = $spfilearray[0];
    $fcontent = $spfile->fread($spfile->getSize());
    $fpath = $spfile->getRealPath();
    $spfilearray[0] = NULL;
    $spfile = NULL;
    $this->fileSystem->delete($fpath);
    $filename = str_replace('"', '', explode("=", explode(";", $spfilearray[2]['Content-Disposition'][0])[1])[1]);
    // Step 5: Import the job for review and accept.
    $this->importJob($job, $fcontent, $filename);
    $this->postImportProcess($request_id, $job_info);
    // If all items are processed and pending count is 0 then send aknowledgment.
    $updateAknowledgeReady = $this->getItemIdFromRequestId($job, $prId, $job_info['requestIds'], $updateId);
    if($updateAknowledgeReady == 0) {
      $this->sendAcknowlegement($job_from, $updateId);
    }
  }

  /**
   * Function to import Job.
   *
   * @param \Drupal\tmgmt\Entity\Job $job
   *   Jobs which nees to import.
   * @param string $filedatastring
   *   File content to process.
   * @param string $name
   *   File name to import the job.
   */
  protected function importJob(Job $job, string $filedatastring, string $name) {
    try {
      $message = '';
      $path = $job->getSetting('scheme') . self::PUBLIC_RECEIVED_FILE_PATH . $name;
      $dirname = dirname($path);
      // If not able to create directory return.
      if (!$this->fileSystem->prepareDirectory($dirname, FileSystemInterface::CREATE_DIRECTORY)) {
        return;
      }
      $file = $this->fileRepository->writeData($filedatastring, $path, FileSystemInterface::EXISTS_REPLACE);
      $mimetype = $file->getMimeType();
      $filearray = NULL;
      if ($mimetype == 'application/zip') {
        $ziparchive = new \ZipArchive();
        $openachresult = $ziparchive->open($this->fileSystem->realpath($path));
        if ($openachresult) {
          $pathtoextract = $job->getSetting('scheme') . self::PUBLIC_RECEIVED_FILE_PATH . pathinfo($path)['filename'];
          if ($ziparchive->extractTo($this->fileSystem->realpath($pathtoextract))) {
            $ziparchive->close();
            $file->delete();
            $filearray = $this->cpcreateFileObjectFromStdClass($this->fileSystem->scanDirectory($pathtoextract, '/.*\.xlf$/'));
          }

        }
      }
      else {
        $filearray = [$file];
      }
      foreach ($filearray as $fileitem) {
        $extension = pathinfo($fileitem->getFileUri(), PATHINFO_EXTENSION);
        $plugin = new Xliff();
        if ($plugin) {
          // Validate the file on job.
          $validated_job = $plugin->validateImport($fileitem->getFileUri(), $job);
          if (!$validated_job) {
            \Drupal::logger('TMGMT_CONTENTAPI')->error('Failed to validate file %FILE for job %JOB, import aborted.', [
              '%JOB' => $job->label(),
              '%FILE' => $fileitem->getFileName(),
            ]);
            $job->addMessage('Failed to validate file, import aborted.', [], 'error');
          }
          elseif ($validated_job->id() != $job->id()) {
            \Drupal::logger('TMGMT_CONTENTAPI')->notice('The imported file job id %file_id does not match the job id %job_id.', [
              '%file_id' => $validated_job->id(),
              '%job_id' => $job->id(),
            ]);
            $job->addMessage('The imported file job id @file_id does not match the job id @job_id.', [
              '@file_id' => $validated_job->id(),
              '@job_id' => $job->id(),
            ], 'error');
          }
          else {
            try {
              // Validation successful, start import.
              // TODO: Check if it makes sense to check only the active state here?

              GeneralHelper::resetJobandItemsToActive($job, $fileitem);

              $job->addTranslatedData($plugin->import($fileitem->getFileUri()));
              \Drupal::logger('TMGMT_CONTENTAPI')->notice('Successfully imported file %file_id for job %job_id.', [
                '%file_id' => $fileitem->getFileName(),
                '%job_id' => $job->label(),
              ]);
              $job->addMessage('File @filename imported successfully', ['@filename' => $fileitem->getFileName()]);
            }
            catch (\Exception $e) {
              \Drupal::logger('TMGMT_CONTENTAPI')->error('File import for job %JOB failed with the following message: %message', [
                '%JOB' => $job->label(),
                '%message' => $e->getMessage(),
              ]);
              $msg = 'File import failed with the following message: ' . $e->getMessage();
              if (strlen($msg) > 200) {
                $msg = substr($msg, 0, 200);
              }
              $job->addMessage($msg, [], 'error');

            }
          }
        }
      }
    }
    catch (\Exception $ep) {
      throw $ep;
    }
  }

  /**
   * Function to cp create file object from stdclass.
   */
  protected function cpcreateFileObjectFromStdClass($filearray) {
    $toreturn = [];
    foreach ($filearray as $fl) {
      $fileMime = $this->guesser->guessMimeType($fl->uri);
      $file = File::create([
      'uid' => \Drupal::currentUser()->id(),
      'filename' => $this->fileSystem->basename($fl->filename),
      'uri' => $fl->uri,
      'filemime' => $fileMime,
      'filesize' => filesize($fl->uri),
      'status' => 1
      ]);
      $file->save();
      array_push($toreturn, $file);
    }
    return $toreturn;
  }

  /**
   * Batch dispatch submission finished callback.
   */
  public static function batchSubmitFinished($success, $results, $operations) {
    if ($success) {
      $translator = [];
      $capi = new TokenApi();
      $translators = \Drupal::entityTypeManager()->getStorage('tmgmt_translator')->loadByProperties(['plugin' => 'contentapi']);
      if ($translators) {
        foreach ($translators as $translatorItem) {
          $translator = $translatorItem;
          break;
        }
      }
      $capisettings = $translator->getSetting('capi-settings');
      $token = $capi->getToken($capisettings['capi_username_ctt'], $capisettings['capi_password_ctt']);

      $requestApi = new RequestApi();

      $jobIds = array_keys($results);
      $localIsAutoAccept = $translator->isAutoAccept();
      foreach($jobIds as $jobId){
        $requests = $requestApi->jobsJobIdRequestsGet($token, $jobId);
        $countCompleted = 0;
        foreach($requests as $req){
          if($req->getStatusCode() == "CANCELLED" || $req->getStatusCode() == "TRANSLATION_APPROVED" || $req->getStatusCode() == "COMPLETED"){
            $countCompleted = $countCompleted + 1;
          }
        }

        $newState = "IN_TRANSLATION";
        $tjid = $results[$jobId];
        $drupalJob = Job::load($tjid);
        if($countCompleted == count($requests)){

          if($drupalJob->getState() != Job::STATE_FINISHED && $localIsAutoAccept){
            $drupalJob->setState(Job::STATE_FINISHED);
            foreach ($drupalJob->getItems() as $item) {
              $item->setState(JobItem::STATE_ACCEPTED);
            }
          }
          $newState = "COMPLETED";
        }
        $updateTime =  new DrupalDateTime('now','UTC');
        $updateTime =  $updateTime->format("Y-m-d H:i:s:u");

        $element_array = [
          'jobId' => $jobId ,
          'jobStatus' => $newState,
          'providerId' => $drupalJob->getSetting("capi-settings")["provider"],
          'jobType' => 2,
          'updateTime' => $updateTime
        ];
        ConentApiHelper::capiJobSetStoredData($drupalJob, $element_array);
      }
      $message = t('The batch processing has been completed successfully.');
    }
    else {
      $error_operation = reset($operations);
      $message = t('An error occurred while processing @operation with arguments : @args', [
        '@operation' => $error_operation[0],
        '@args' => print_r($error_operation[0], TRUE),
      ]);
    }
    \Drupal::messenger()->addMessage($message);
  }


  /**
   * Function to approve item in db, as well as capi.
   *
   * @param array $job_info
   *   Job for which this status needs to be updated.
   */
  protected function approveItemStatusToCapiAndConnector($type, $prId, $request_id) {
    $this->arrayOfRequestId = new ArrayOfRequestIds(['request_ids' => $request_id]);
    // Verify if all job items are imported.
    // If yes then acknowledge the jobs.
    // Else do no aknowledge Job.
    // If autoaccept set then process the job and send response as TRANSLATION_APPROVED.
    if ($this->isAutoAccept) {
      if ($type == 1) {
        $returnarray = $this->requestApi->jobsJobIdRequestsApprovePutCtt($this->cttToken, $prId, $this->arrayOfRequestId);
      }
      else {
        $returnarray = $this->requestApi->jobsJobIdRequestsApprovePut($this->capiToken, $prId, $this->arrayOfRequestId);
      }
      if (count($returnarray) == 1) {
        \Drupal::logger('Item import and acknowledge')->notice('APPROVE STATUAS UPDATE CAPI - status for jobid: @prId and request id: @requestid', [
          '@prId' => $prId,
          '@requestid' => json_encode($request_id),
        ]);
        } else {
        \Drupal::logger('Item import and acknowledge failed')->error('APPROVE STATUAS UPDATE CAPI- Failed to update status job for jobid: @prId and request id: @requestid', [
          '@prId' => $prId,
          '@requestid' => json_encode($request_id),
        ]);
      }
    }
  }

  /**
   * Approves the job status.
   *
   * @param mixed $job The job object.
   * @param int $prId The project ID.
   * @param string $job_from The job source.
   * @param int $providerId The provider ID.
   * @return void
   */
  public function approveJobStatus($job, $prId, $job_from, $providerId) {
    // In Last step we have send status "TRANSLATION_APPROVED" to provider.
    // We also need to make sure same status get updated in connector side.
    // JIRA : 1195821 : Job status column ambuguity.
    // This needs to set if only if all items has been processed for the job.
    // Check if all items are processed for this JOB, then set this as true.
    $isPendingItems = $job->getCountPending();
    if( $this->isAutoAccept && !$isPendingItems) {
      $jobstatus = 'TRANSLATION_APPROVED';
      $updateTime = $this->drupalDateTime->format("Y-m-d H:i:s:u");
      $element_array = [
        'jobId' => $prId ,
        'jobStatus' => $jobstatus,
        'providerId' => $providerId,
        'jobType' => $job_from,
        'updateTime' => $updateTime
      ];
      ConentApiHelper::capiJobSetStoredData($job, $element_array);
      \Drupal::logger('Job finished')->notice('jobid: @prId and request id: @providerId and PendingCount: @count', [
        '@prId' => $prId,
        '@providerId' => $providerId,
        '@count' => $isPendingItems,
      ]);
    }
  }

  /**
   * Process the job after importing.
   *
   * @param int $request_id The ID of the request.
   * @param array $job_info The job information.
   * @return void
   */
  public function postImportProcess($request_id, $job_info) {
    // Once a batch of items get process for a job
    // Update the status af process items in capi.
    $this->approveItemStatusToCapiAndConnector($job_info["type"], $job_info['jobId'], [$request_id]);
    // Once all Items of a job get processed then Mark it as Approved
    // In connectors DB.
    $this->approveJobStatus($job_info['drupalJob'], $job_info['jobId'], $job_info["type"], $job_info["providerId"]);
  }

  /**
   * Retrieves the item ID from the request ID.
   *
   * @param Job $job The job object.
   *
   * @param string $prId The job id.
   *
   * @param array $requestIds The array of request IDs.
   *
   * @return mixed The item ID associated with the request ID.
   */
  protected function getItemIdFromRequestId(Job $job, string $prId, array $requestIds, $updateId) {
    //Get all pendings items ids from the job .
    $pending_items = $this->getPendingItemsForJob($job);
    // Check if the items associated with request ID from this Update is in the pending state.
    $item_ids_in_update = [];
    // Get item id associated with request id, Capi-remote object holds details of the job.
    // Which will give us the request id and item id mapping.
    $submittedrequestsarray = unserialize($job->getSetting('capi-remote'));
    $arraywithrequest = $submittedrequestsarray[0];
    // Loop through the array to get the request id and item id mapping.
    foreach($submittedrequestsarray as $arraywithrequest) {
      foreach ($arraywithrequest as $request) {
        // Fetch request id from the request object.
        $request_id = $request->getRequestId();
        // If request id is not in the request id array then continue.
        if(!in_array($request_id, $requestIds)) {
          continue;
        }
        // Fetch item id from the request object.
        $requestSourceNativeId = explode("_", $request->getSourceNativeId())[1];
        // Add item id to the array.
        $item_ids_in_update[$request_id] = $requestSourceNativeId;
      }
    }
    // Get count, if any item from this update is still in pending state.
    $is_pending_item_from_update = count(array_intersect($item_ids_in_update, $pending_items));
    return $is_pending_item_from_update;
  }

  /**
   * Retrieves the pending items for a given job.
   *
   * @param Job $job The job for which to retrieve pending items.
   * @return array An array of pending items for the job.
   */
  public function getPendingItemsForJob(Job $job) {
    $pending_items = [];
    $query = \Drupal::entityQuery('tmgmt_job_item')
      ->accessCheck(TRUE)
      ->condition('tjid', $job->id())
      ->condition('state', 1)
      ->sort('tjiid', 'DESC');
    $result = $query->execute();
    foreach($result as $item) {
      $pending_items[] = $item;
    }
    return $pending_items;
  }

  /**
   * Function to capijob from drupal job.
   *
   * @param \Drupal\tmgmt\Entity\Job $job
   *   Jobs which nees to import.
   */
  public function getCapiJobFromDrupalJob(Job $job) {
    $prId = GeneralHelper::getCpJobIdfromLocJob($job);
    $find_job = $this->jobApi->FindJob($this->capiToken, $prId);
    if ($find_job != 200) {
      $requestsarray = $this->requestApi->jobsJobIdRequestsGetCtt($this->cttToken, $prId);
      $projectInfo = $this->jobApi->jobsJobIdGetCtt($this->cttToken, $prId, "fullWithStats");
      $jobstatus = $projectInfo->getStatusCode()->getStatusCode();
    }
    else {
      $requestsarray = $this->requestApi->jobsJobIdRequestsGet($this->capiToken, $prId);
      $projectInfo = $this->jobApi->jobsJobIdGet($this->capiToken, $prId, "fullWithStats");
      if ($projectInfo->getStatusCode() != 'COMPLETED') {
        $jobstatus = $this->jobApi->jobsJobIdGetStatus($this->capiToken, $prId);
      }
      else {
        $jobstatus = $projectInfo->getStatusCode();
      }
    }
    $requestIds = [];
    foreach ($requestsarray as $request) {
      array_push($requestIds, $request->getRequestId());
    }
    $updateTime = $this->drupalDateTime->format("Y-m-d H:i:s:u");
    $jobInfo['type'] = ($find_job != 200 ? 1 : 2);
    $jobInfo['jobId'] = $projectInfo->getJobId();
    $jobInfo['updateId'] = '';
    $jobInfo['requestIds'] = $requestIds;
    $jobInfo['providerId'] = $projectInfo->getProviderId();
    $jobInfo['status'] = $jobstatus;
    $jobInfo['drupalJob'] = $job;
    $jobInfo['updateTime'] = $updateTime;
    return $jobInfo;
  }

  /**
   * Function to import job via xlf file.
   *
   * @param \Drupal\tmgmt\Entity\Job $job
   *   Jobs which nees to import.
   */
  public function importJobViaFile(Job $job) {
    if ($file = file_save_upload('file', ['file_validate_extensions' => ['xlf']], FALSE, 0)) {
      $extension = pathinfo($file->getFileUri(), PATHINFO_EXTENSION);
      $plugin = new Xliff();
      if ($plugin) {
        // Validate the file on job.
        $validated_job = $plugin->validateImport($file->getFileUri(), $job);
        if (!$validated_job) {
          $job->addMessage('Failed to validate file, import aborted.', [], 'error');
        }
        elseif ($validated_job->id() != $job->id()) {
          $job->addMessage('The imported file job id @file_id does not match the job id @job_id.', [
            '@file_id' => $validated_job->id(),
            '@job_id' => $job->id(),
          ], 'error');
        }
        else {
          try {
            // Validation successful, start import.
            // Check if it makes sense to check only the active state here?
            GeneralHelper::resetJobandItemsToActive($job, $file);
            $job->addTranslatedData($plugin->import($file->getFileUri()));
            $job->addMessage('Successfully imported file.');
            ConentApiHelper::checkJobFinishAndApproveRemote($job);
          }
          catch (\Exception $e) {

            $msg = 'File import failed with the following message: ' . $e->getMessage();
            if (strlen($msg) > 200) {
              $msg = substr($msg, 0, 200);
            }
            $job->addMessage($msg, [], 'error');
          }
        }
        tmgmt_write_request_messages($job);
      }
    }

  }

}
